#define PWMPin 10

float frq = 1000.0; // 0.238Hz <= frq <= 8MHz
float duty = 0.5; // 指定したいデューティ比

void pulse() {
  pinMode(PWMPin, OUTPUT);

// TCCR1A: COM1A1, COM1A0, COM1B1, COM1B0, -, -, WGM11, 1GM10
// OC1A set on compare match, clear at BOTTOM (COM1A = 0b01)
// OC1B clear on compare match, set at BOTTOM (COM1B = 0b10)
// OCR1A Fast PWM Mode, TOP値=OCR1A (WGM11:10 = 0b11)
  TCCR1A = _BV(COM1A0) | _BV(COM1B1) | _BV(WGM11) | _BV(WGM10); // Fast PWM mode - compare to OCR1A

// TCCR1B: ICNC1, ICES1, -, WGM13, WGM12, CS12, CS11, CS10
// OCR1A Fast PWM Mode (WGM13:12 = 0b11)
// CS12:10 001:ck/1, 010:ck/8, 011:ck/64, 100:ck/256, 101:ck/1024
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10); // 001:ck/1 仮設定
//  TCCR1B = 0b00011001;  // ck/1
//  TCCR1B = 0b00011010;  // ck/8
//  TCCR1B = 0b00011011;  // ck/64
//  TCCR1B = 0b00011100;  // ck/256
//  TCCR1B = 0b00011101;  // ck/1024
  TCNT1 = 0x0000;               // TCNT1初期化
//  OCR1A = 2000;                 // パルス周期1ms 仮設定
//  OCR1B = 1000;                 // パルス幅.5ms 仮設定

  float divide;
  int idiv = calcDivide(frq);
  if (idiv == 0) {
    frq = 8000000.0;  // up to 8MHz
    divide = 1.0;     // ck/1
    setCounter(1);
  } else if (idiv == 1025) {
    frq = 0.2384185791015625;     // down to 0.238Hz
    divide = 1024.0;  // ck/1024
    setCounter(1024);
  } else {
    divide = (float) calcDivide(frq);
    setCounter(idiv);
  }
  // TOP値指定
  OCR1A = (unsigned int)(16000000.0 / frq / divide - 1);
  // Duty比指定
  OCR1B = (unsigned int)(16000000.0 / frq / divide * duty - 1);
}

void setCounter(int divide) {
  if (divide == 1) {
    TCCR1B = 0b00011001;  // ck/1
  } else if (divide == 8) {
    TCCR1B = 0b00011010;  // ck/8
  } else if (divide == 64) {
    TCCR1B = 0b00011011;  // ck/64
  } else if (divide == 256) {
    TCCR1B = 0b00011100;  // ck/256
  } else if (divide == 1024) {
    TCCR1B = 0b00011101;  // ck/1024
  } else {
    TCCR1B = 0b00011001;  // ck/1
  }
}

int calcDivide(float freq) {
  if ((unsigned long)(16000000.0 / frq - 1) < 1) {              // too high > 8MHz
    return (0);
  } else if ((unsigned long)(16000000.0 / frq - 1) < 65536L) {  // ck/1 > 244.1Hz
    return (1);
  } else if ((unsigned long)(2000000.0 / frq - 1) < 65536L) {   // ck/8 > 30.5Hz
    return (8);
  } else if ((unsigned long)(250000.0 / frq - 1) < 65536L) {    // ck/64 > 3.81Hz
    return (64);
  } else if ((unsigned long)(62500.0 / frq - 1) < 65536L) {     // ck/256 > 0.954Hz
    return (256);
  } else if ((unsigned long)(15625.0 / frq - 1) < 65536L) {     // ck/1024 > 0.238Hz
    return (1024);
  } else {                                                      // too low < 0.238Hz
    return (1025);
  }
}
